/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)ioctl.c	1.4	99/02/18 SMI"

#include	<stdio.h>
#include	<sys/termio.h>
#include	<termios.h>
#include	<unistd.h>
#include	<string.h>
#include	<errno.h>
#include	<sys/socket.h>
#include	<fcntl.h>
#if defined(UNIXWARE) || defined(GEMINI) || defined(SOLARIS)
# include	<sys/filio.h>
# include	<sys/sockio.h>
# include	<net/if.h>
#else
# include	<sys/net/if.h>
#endif
#if defined(OSR5)
# include	<sys/vtkd.h>
# include	<sys/console.h>
# include	<sys/vid.h>
#else
#endif
#include	"lxrun.h"
#include	"linux/socket.h"
#include	"linux/termios.h"
#include	"linux/vtkd.h"
#include	"devmap.h"

#ifndef TCGETS
# define TCGETS		XCGETA
# define TCSETS		XCSETA
# define TCSETSW	XCSETAW
# define TCSETSF	XCSETAF
#endif

int tioclinux(int fd, void *arg);

/*
 * ioctl() related system calls
 */

/*
 *	ioctl 		(PARTLY IMPLEMENTED)
 */


int lx_ioctl_default();
int lx_ioctl_audio();


/*
** Initialize the device-to-ioctl-handler map
*/
void
devmap_init(void)
{
	int i;

	for(i=0; i<NMAJ; i++)
		devmap_add_ioctl_handler(lx_ioctl_default, i);


	/* *** PUT REGISTRATION FOR NEW IOCTL HANDLERS HERE *** */

	devmap_add_ioctl_handler(lx_ioctl_audio, devmap_major("/dev/dsp"));
}



int
lx_ioctl_default(int fd, int lx_cmd, unsigned long lx_arg, int maj, int min)
{
	int		cmd;
	int		argtype;

	switch (lx_cmd)
	{
		case LX_TCGETS: 
		{
			int rv ;
			struct termios tio ;
			struct lx_termios* lx_tio=(struct lx_termios *)lx_arg;

			if(!verify_write(lx_tio, sizeof(*lx_tio)))
				/* verify_write() set errno to EFAULT */
				return -1;

			rv = ioctl(fd, TCGETS, &tio);
			if (rv) return rv ;
			memset(lx_tio, 0, sizeof(struct lx_termios));

			/* This isn't really safe, since each side has some
			   flags the other doesn't.  But the cost of doing
			   a bitwise test for each one it just too horrible
			   for words. */
			lx_tio->c_iflag = tio.c_iflag ;
			lx_tio->c_oflag = tio.c_oflag ;
			lx_tio->c_cflag = tio.c_cflag ;
			lx_tio->c_lflag = tio.c_lflag ;

			if (tio.c_lflag & ICANON)
			{
				lx_tio->c_cc[LX_VEOF]	= tio.c_cc[VEOF] ;
				lx_tio->c_cc[LX_VEOL]	= tio.c_cc[VEOL] ;
			}
			else 
			{
				lx_tio->c_cc[LX_VMIN]	= tio.c_cc[VMIN] ;
				lx_tio->c_cc[LX_VTIME]	= tio.c_cc[VTIME] ;
			}

			lx_tio->c_cc[LX_VINTR] 	= tio.c_cc[VINTR] ;
			lx_tio->c_cc[LX_VQUIT] 	= tio.c_cc[VQUIT] ;
			lx_tio->c_cc[LX_VERASE]	= tio.c_cc[VERASE] ;
			lx_tio->c_cc[LX_VKILL]	= tio.c_cc[VKILL] ;
			lx_tio->c_cc[LX_VSWTC]	= tio.c_cc[VSWTCH] ;
			lx_tio->c_cc[LX_VSTART]	= tio.c_cc[VSTART] ;
			lx_tio->c_cc[LX_VSTOP]	= tio.c_cc[VSTOP] ;
			lx_tio->c_cc[LX_VSUSP]	= tio.c_cc[VSUSP] ;
			lx_tio->c_cc[LX_VEOL]	= tio.c_cc[VEOL] ;
			return 0 ;
		}

		case LX_TCSETS:
		case LX_TCSETSW:
		case LX_TCSETSF:
		{
			struct termios tio ;
			struct lx_termios* lx_tio=(struct lx_termios *)lx_arg;
			long sco_cmd = -1;

			if (!verify_read(lx_tio, sizeof(*lx_tio)))
				/* verify_read() set errno to EFAULT */
				return -1;

			switch (lx_cmd) 
			{
				case LX_TCSETS:  sco_cmd = TCSETS;	break;
				case LX_TCSETSW: sco_cmd = TCSETSW;	break;
				case LX_TCSETSF: sco_cmd = TCSETSF;	break;
			}

			memset(&tio, 0, sizeof(struct termios)); 

			/* This isn't really safe, since each side has some
			   flags the other doesn't.  But the cost of doing
			   a bitwise test for each one it just too horrible
			   for words. */
			tio.c_iflag = lx_tio->c_iflag ;
			tio.c_oflag = lx_tio->c_oflag ;
			tio.c_cflag = lx_tio->c_cflag ;
			tio.c_lflag = lx_tio->c_lflag ;


			/* On the SCO side, these overlap (sigh) */

			if (lx_tio->c_lflag & LX_ICANON) 
			{
				tio.c_cc[VEOF]	= lx_tio->c_cc[LX_VEOF] ;
				tio.c_cc[VEOL]	= lx_tio->c_cc[LX_VEOL] ;
			}
			else
			{
				tio.c_cc[VTIME]	= lx_tio->c_cc[LX_VTIME] ;
				tio.c_cc[VMIN]	= lx_tio->c_cc[LX_VMIN] ;
			}

			tio.c_cc[VINTR]  = lx_tio->c_cc[LX_VINTR] ;
			tio.c_cc[VQUIT]  = lx_tio->c_cc[LX_VQUIT] ;
			tio.c_cc[VERASE] = lx_tio->c_cc[LX_VERASE] ;
			tio.c_cc[VKILL]	 = lx_tio->c_cc[LX_VKILL] ;
			tio.c_cc[VSWTCH] = lx_tio->c_cc[LX_VSWTC] ;
			tio.c_cc[VSTART] = lx_tio->c_cc[LX_VSTART] ;
			tio.c_cc[VSTOP]  = lx_tio->c_cc[LX_VSTOP] ;
			tio.c_cc[VSUSP]  = lx_tio->c_cc[LX_VSUSP] ;

			return ioctl(fd, sco_cmd, &tio);
		}

		case LX_TCGETA:
		       {
		       int rv ;
		       struct termio tio ;
		       struct lx_termio* lx_tio = (struct lx_termio *) lx_arg ;
		       rv = ioctl(fd, TCGETA, &tio);
		       if (rv) return rv ;

		       if (! verify_write(lx_tio, sizeof(*lx_tio)))
			       return -1; /* verify_write() set errno to EFAULT */

		       memset(lx_tio, 0, sizeof(struct lx_termio));

		       /* This isn't really safe, since each side has some
			  flags the other doesn't.  But the cost of doing
			  a bitwise test for each one it just too horrible
			  for words. */
		       lx_tio->c_iflag = tio.c_iflag ;
		       lx_tio->c_oflag = tio.c_oflag ;
		       lx_tio->c_cflag = tio.c_cflag ;
		       lx_tio->c_lflag = tio.c_lflag ;


		       if (tio.c_lflag & ICANON)
			       {
			       lx_tio->c_cc[LX_VEOF]   = tio.c_cc[VEOF] ;
			       lx_tio->c_cc[LX_VEOL]   = tio.c_cc[VEOL] ;
			       }
		       else
			       {
			       lx_tio->c_cc[LX_VMIN]   = tio.c_cc[VMIN] ;
			       lx_tio->c_cc[LX_VTIME]  = tio.c_cc[VTIME] ;
			       }
		       lx_tio->c_cc[LX_VINTR]  = tio.c_cc[VINTR] ;
		       lx_tio->c_cc[LX_VQUIT]  = tio.c_cc[VQUIT] ;
		       lx_tio->c_cc[LX_VERASE] = tio.c_cc[VERASE] ;
		       lx_tio->c_cc[LX_VKILL]  = tio.c_cc[VKILL] ;
		       lx_tio->c_cc[LX_VSWTC]  = tio.c_cc[VSWTCH] ;
		       lx_tio->c_cc[LX_VSTART] = tio.c_cc[VSTART] ;
		       lx_tio->c_cc[LX_VSTOP]  = tio.c_cc[VSTOP] ;
		       lx_tio->c_cc[LX_VSUSP]  = tio.c_cc[VSUSP] ;
		       lx_tio->c_cc[LX_VEOL]   = tio.c_cc[VEOL] ;
		       return 0 ;
		       }

		case LX_TCSETA:
		case LX_TCSETAW:
		case LX_TCSETAF:
		       {
		       struct termio tio ;
		       struct lx_termio* lx_tio = (struct lx_termio *) lx_arg ;
		       long sco_cmd = -1 ;

		       if (! verify_read(lx_tio, sizeof(*lx_tio)))
			       return -1;

		       switch (lx_cmd)
			       {
			       case LX_TCSETA:  sco_cmd = TCSETA ;  break;
			       case LX_TCSETAW: sco_cmd = TCSETAW ; break;
			       case LX_TCSETAF: sco_cmd = TCSETAF ; break;
			       }

		       memset(&tio, 0, sizeof(struct termio));

		       /* This isn't really safe, since each side has some
			  flags the other doesn't.  But the cost of doing
			  a bitwise test for each one it just too horrible
			  for words. */
		       tio.c_iflag = lx_tio->c_iflag ;
		       tio.c_oflag = lx_tio->c_oflag ;
		       tio.c_cflag = lx_tio->c_cflag ;
		       tio.c_lflag = lx_tio->c_lflag ;


		       /* On the SCO side, these overlap (sigh) */

		       if (lx_tio->c_lflag & LX_ICANON)
			       {
			       tio.c_cc[VEOF]  = lx_tio->c_cc[LX_VEOF] ;
			       tio.c_cc[VEOL]  = lx_tio->c_cc[LX_VEOL] ;
			       }
		       else
			       {
			       tio.c_cc[VTIME] = lx_tio->c_cc[LX_VTIME] ;
			       tio.c_cc[VMIN]  = lx_tio->c_cc[LX_VMIN] ;
			       }

		       tio.c_cc[VINTR]	= lx_tio->c_cc[LX_VINTR] ;
		       tio.c_cc[VQUIT]	= lx_tio->c_cc[LX_VQUIT] ;
		       tio.c_cc[VERASE]	= lx_tio->c_cc[LX_VERASE] ;
		       tio.c_cc[VKILL]	= lx_tio->c_cc[LX_VKILL] ;
		       tio.c_cc[VSWTCH]	= lx_tio->c_cc[LX_VSWTC] ;
		       tio.c_cc[VSTART]	= lx_tio->c_cc[LX_VSTART] ;
		       tio.c_cc[VSTOP]	= lx_tio->c_cc[LX_VSTOP] ;
		       tio.c_cc[VSUSP]	= lx_tio->c_cc[LX_VSUSP] ;

		       return ioctl(fd, sco_cmd, &tio);
		       }

		case LX_TCSBRK:
		case LX_TCXONC:
		case LX_TCFLSH:
		case LX_TIOCEXCL:
		case LX_TIOCNXCL: {
			char msg[32];
			sprintf(msg, "ioctl(0x%x)", lx_cmd);
			return enosys(msg);
		}

		case LX_TIOCGPGRP:
			{
			pid_t pgrp=tcgetpgrp(fd);
			if(pgrp == -1)
				return -1;
			*((pid_t *)lx_arg) = pgrp;
			return 0;
			}

		case LX_TIOCSPGRP:
			return tcsetpgrp(fd, *((pid_t *)lx_arg));

		case LX_TIOCGWINSZ:
			return ioctl(fd, TIOCGWINSZ, lx_arg);

		case LX_TIOCSWINSZ:
			return enosys("ioctl(TIOCSWINSZ)");

		case LX_FIONREAD:
			return ioctl(fd, FIONREAD, lx_arg);
			
		case LX_FIONBIO:
			return ioctl(fd, FIONBIO, lx_arg);

		       /* It's a gift that the structures seem to
			  match "well enough" */
		case LX_SIOCGIFCONF:
			return ioctl(fd, SIOCGIFCONF, lx_arg);

		case LX_SIOCGIFFLAGS:
			return ioctl(fd, SIOCGIFFLAGS, lx_arg);
 
		case LX_SIOCGIFADDR:
			return ioctl(fd, SIOCGIFADDR, lx_arg);
 
		case LX_SIOCGIFNETMASK:
			return ioctl(fd, SIOCGIFNETMASK, lx_arg);
 
		case LX_SIOCGIFBRDADDR:
			return ioctl(fd, SIOCGIFBRDADDR, lx_arg);

		case LX_TIOCLINUX:
#ifdef OSR5
			return tioclinux(fd, (void *)lx_arg);
#else
			return enosys("ioctl(TIOCLINUX)");
#endif

#ifdef OSR5
		case LX_KDGETMODE:
			return ioctl(fd, KDGETMODE, lx_arg);

		case LX_KDSETMODE:
			return ioctl(fd, KDSETMODE, lx_arg);

		case LX_VT_GETSTATE:	/* partially implemented */
			{
			struct lx_vt_stat *stat = (struct lx_vt_stat *)lx_arg;
			struct vid_info vi;
			int tmpfd;

			/* 16-bit bitmask telling us which multiscreens are
			** usable.  We'll assume the default OSR5 configuration
			** -- i.e., we have 12 active multiscreens.  */
			stat->v_state = 0x0fff;

			/* # of active multiscreen (1-based) */
			/*
			** We have a slight problem here in that we must
			** be root to open /dev/vga -- necessary to determine
			** the current multiscreen -- and on Linux, VT_GETSTATE
			** is not a privileged ioctl.  Oh well.
			**
			** Also, we're assuming that everyone these days has
			** vga.  Technically, we should be doing a CONS_CURRENT
			** to find out what kind of display we have before
			** blindly opening /dev/vga.
			*/
			if((tmpfd = open("/dev/vga", O_RDONLY)) == -1)
				return -1;
			vi.size = sizeof(struct vid_info);
			ioctl(tmpfd, CONS_GETINFO, &vi);
			close(tmpfd);
			stat->v_active = vi.m_num+1;
				
			/* ??? */
			/* stat->v_signal = */

			}
#else
		case LX_KDGETMODE:	
		case LX_KDSETMODE:
		case LX_VT_GETSTATE: {
			char msg[32];
			sprintf(msg, "ioctl(0x%x)", lx_cmd);
			return enosys(msg);
		}
#endif
		case LX_TIOCSCTTY:
#if defined(SOLARIS)
			/*
			 * SysV semantics for controlling terminals say that
			 * the first terminal device not already associated
			 * with a session, and not opened with O_NOCTTY, will
			 * become the controlling tty for the session.
			 *
			 * Presumably, if a process were calling TIOCSCTTY
			 * on a file descriptor, it will have opened it
			 * without O_NOCTTY, so it will have already become
			 * the controlling tty, and 0 would be the correct
			 * return value.
			 * 
			 * XXX: This is incorrect if a process creates a new
			 * session, opens two terminal devices, and wants
			 * the second to be its controlling terminal.
			 * That's pretty hard to deal with.
			 */
			return 0;
#else
			/*FALLTHROUGH*/
#endif
		case LX_TIOCSETD:
		case LX_TIOCGETD:
		case LX_FIONCLEX:
		case LX_FIOCLEX:
		case LX_FIOASYNC:
		case LX_TIOCOUTQ:
		case LX_TIOCSTI:
		case LX_TIOCMGET:
		case LX_TIOCMBIS:
		case LX_TIOCMBIC:
		case LX_TIOCMSET:
		case LX_TIOCGSOFTCAR:
		case LX_TIOCSSOFTCAR:
		case LX_TIOCCONS:
		case LX_TIOCGSERIAL:
		case LX_TIOCSSERIAL:
		case LX_TIOCPKT:
		case LX_TIOCNOTTY:
		case LX_TCSBRKP:
		case LX_TIOCTTYGSTRUCT:
		case LX_TIOCSERCONFIG:
		case LX_TIOCSERGWILD:
		case LX_TIOCSERSWILD:
		case LX_TIOCGLCKTRMIOS:
		case LX_TIOCSLCKTRMIOS:
		case LX_TIOCSERGSTRUCT:
		case LX_TIOCSERGETLSR:
		case LX_TIOCSERGETMULTI:
		case LX_TIOCSERSETMULTI: {
			char msg[32];
			sprintf(msg, "ioctl(0x%x)", lx_cmd);
			return enosys(msg);
		}

		default: {
			char msg[64];
			sprintf(msg, "ioctl(0x%x) [unlisted]", lx_cmd);
			return enosys(msg);
		}
	}
}

/*
** ioctl mapping for OSS audio drivers
*/
int
lx_ioctl_audio(int fd, int lx_cmd, unsigned long lx_arg, int maj, int min)
{
	unsigned char	cmdkey = (lx_cmd >> 8) & 0xff;

	if(cmdkey == 'P' ||	/* DSP */
	   cmdkey == 'M' ||	/* Mixer */
	   cmdkey == 'm' ||	/* MIDI */
	   cmdkey == 'C' ||	/* Co-Processor */
	   cmdkey == 'T' ||	/* Timer */
	   cmdkey == 'Q')	/* Sequencer */
		{
		/*
		** Since SCO audio is essentially the same as
		** Linux audio, there's not much conversion to
		** do.  (There wouldn't be any, but the direction
		** information is handled differently.)
		*/
		unsigned int sco_cmd, dirbits;

		/*
		** TTD: This should be changed so we use macros from
		** ioctl.h and/or soundcard.h instead of using hard-
		** coded values.
		*/

		/* Collect direction bits from original command */
		dirbits = lx_cmd & 0xc0000000;

		/* Perform transformation */
		if(dirbits & 0x80000000)
			{
			dirbits &= ~0x80000000;
			dirbits |=  0x20000000;
			}

		/* Mask them back in */
		sco_cmd = (lx_cmd & 0x1ffffff) | dirbits;

		/* Call the audio ioctl */
		return ioctl(fd, sco_cmd, lx_arg);
		}

	errno=ENOSYS;
	return -1;
}



#ifdef OSR5

/*
** Emulation of TIOCLINUX ioctl (some subcodes) for SCO OpenServer 5.0.x
**
** Sep  6 1998	- Initial revision (Steve Ginzburg)
*/

#include <sys/console.h>
#include <stdlib.h>

char sel_buffer[256];
int sel_count=-1;

typedef struct region {
	short xs, ys, xe, ye;	/* Selection boundaries, 0-based */
	short sel_mode;		/* 0 = char, 1 = word, 2 = line */
} Region;


Region hilited;


void toggle_highlight(Region *r)
{
	int start, end, i;
	char *disp;			/* display memory */

	disp=(char *)ioctl(0, MAPCONS, 0L);

	if(r->sel_mode == 1)		/* word-based selection */
		{
		/* Not implemented.  Default to char-based. */
		}
	else if(r->sel_mode == 2)	/* line-based selection */
		{
		r->xs = 0; r->xe = 79;
		}

	start = r->xs + 80*r->ys;
	end = r->xe + 80*r->ye;

	/* Guarantee start <= end */
	if(start > end)
		start ^= end ^= start ^= end;

	sel_count = end-start;
	if(sel_count > sizeof sel_buffer)
		sel_count = sizeof sel_buffer;

	for(i=0; i<=sel_count && i<sizeof sel_buffer; ++i)
		{
		disp[(i+start)*2+1] ^= 0x50;
		sel_buffer[i] = disp[(i+start)*2];
		}
}



int tioclinux(int fd, void *arg)
{
	switch(*((char *)arg))
		{
		case 2:		/* copy */
			{
			Region r;
			memcpy(&r, ((char *)arg)+1, sizeof(Region));
			--r.xs; --r.ys; --r.xe; --r.ye;

			/* unhighlight previous region */
			if(sel_count != -1)
				toggle_highlight(&hilited);

			memcpy(&hilited, &r, sizeof(Region));

			toggle_highlight(&hilited);

			return 0;
			}

		case 3:		/* paste */
			return write(fd, sel_buffer, sel_count);
			/*
			** This is not correct!  Is there a way to emulate
			** this properly on OpenServer?  I don't know...
			*/

		case 5:		/* define word boundaries */
			/* not implemented, but pretend it is */
			return 0;

		case 6:		/* get state of shift key */
			/* I don't know how to do this on OSR5.  So we'll   **
			** just tell the app that shift is not pressed.     */
			*(char *)arg = 0;
			return 0;

		default:
			errno=ENOSYS; return -1;
		}
}

#endif
